/*
 * $Id: Mc.java,v 1.3 2002/07/30 04:01:04 kenta Exp $
 *
 * Copyright (c) 2002 Kenta Cho. All rights reserved.
 */

import com.nttdocomo.ui.*;
import com.nttdocomo.opt.ui.j3d2.*;
import java.util.Random;

/**
 * Laser command main class.
 *
 * @author cs8k-cyu@asahi-net.or.jp
 * @version $Revision: 1.3 $
 */
public class Mc extends IApplication {
  private NullCanvas canvas;
  private C3DGraphics g;
  private int scWidth, scHeight;

  private final int fx1 = 65536;

  private boolean backLight = true;
  /**
   * Handle the backlight.
   */
  private void setBackLight() {
    int blStat;
    if ( backLight ) {
      blStat = PhoneSystem.ATTR_BACKLIGHT_ON;
    } else {
      blStat = PhoneSystem.ATTR_BACKLIGHT_OFF;
    }
    PhoneSystem.setAttribute(PhoneSystem.DEV_BACKLIGHT, blStat);
  }

  /**
   * Initialize.
   */
  private void init() {
    canvas = new NullCanvas();
    canvas.setSoftLabel(Frame.SOFT_KEY_2, "Exit");
    scWidth = canvas.getWidth();
    scHeight = canvas.getHeight();
    setBackLight();
    Display.setCurrent(canvas);
    initTitle();
  }

  /**
   * Init a canvas after the C3DGraphics instance becomes available.
   */
  private void initCanvas() {
    g.set3DMode(true);

    g.viewport(0, 0, scWidth, scHeight);
    g.clearColor(0, 0, 0, fx1);

    g.enable(g.C3DL_TEXTURE_2D);
    makeTexture();

    g.blendFunc(g.C3DL_SRC_ALPHA, g.C3DL_ONE);
    g.enable(g.C3DL_BLEND);
    g.enable(g.C3DL_LINE_SMOOTH);
    g.disable(g.C3DL_DEPTH_TEST);
    g.disable(g.C3DL_CULL_FACE);

    g.matrixMode(g.C3DL_PROJECTION);
    g.loadIdentity();
    int aspect = fx1 * scWidth / scHeight;
    g.perspective(45*fx1, aspect, fx1/256, 512*fx1);

    g.matrixMode(g.C3DL_MODELVIEW);
    g.loadIdentity();

    g.disable(g.C3DL_LIGHTING);
  }

  private final short[] lt = {
    0x7bed,0x6bee,0x7927,0x6b6e,0x79a7,0x79a4,0x796f,0x5bed,0x2492,0x126f,0x5bad,0x4927,0x5fed,0x5ffd,0x7b6f,0x7be4,0x7b7f,0x6bad,0x79cf,0x7492,0x5b6f,0x5b6a,0x5ffa,0x5aad,0x5bd2,0x72a7,0x7b6f,0x2492,0x73e7,0x73cf,0x5bc9,0x79cf,0x79ef,0x7249,0x7bef,0x7bcf,0x3493,0x6496,
  };

  /**
   * Make a texuture with the text.
   */
  private void makeStringTexture(int tn, String s, boolean r) {
    int w, l = s.length();
    if ( l == 1 ) {
      w = 4;
    } else {
      w = 32;
    }
    byte[] bp = new byte[w*8*4];
    char c;
    short p, m;
    int k;
    byte rc, gc, bc;
    for ( int i=0, x=0 ; i<l ; i++, x+=4 ) {
      c = s.charAt(i);
      rc = gc = bc = (byte)255;
      if ( c == ' ' ) continue;
      else if ( c == '[' ) {
        c = (char)('9'-'0'+1 + 'z'-'a'+1);
        gc = bc = (byte)128;
      } else if ( c == ']' ) {
        c = (char)('9'-'0'+1 + 'z'-'a'+1 + 1);
        gc = bc = (byte)128;
      }
      else if ( c >= 'a' && c <='z' ) {
        c -= 'a';
      } else {
        c = (char)(c-'0' + 'z'-'a'+1);
        bc = (byte)128;
      }
      p = lt[c];
      m = 0x4000;
      for ( int j=0 ; j<15 ; j++ ) {
        k = ((j%3)+x + (j/3)*w) * 4;
        if ( r ) k += (8-l)*4*4;
        if ( (p&m) != 0 ) {
          bp[k] = rc; bp[k+1] = gc; bp[k+2] = bc;
          bp[k+3] = (byte)255;
        } else {
          bp[k] = bp[k+1] = bp[k+2] = bp[k+3] = (byte)0;
        }
        m >>= 1;
      }
    }
    g.bindTexture(g.C3DL_TEXTURE_2D, tn);
    g.texImage2D(g.C3DL_TEXTURE_2D, 0,
                 g.C3DL_RGBA8, w, 8, 0, g.C3DL_RGBA, g.C3DL_UNSIGNED_BYTE,
                 bp);
  }

  /**
   * Make images for textures.
   */
  private void makeImageTexture(int idx, String s, int w) {
    byte[] bp = new byte[w*w*4];
    for ( int y=0 ; y<w ; y++ ) {
      for ( int x=0 ; x<w ; x++ ) {
        int i = (x+y*w)*4, j = (x+y*w)*3;
        bp[i]   = (byte)s.charAt(j);
        bp[i+1] = (byte)s.charAt(j+1);
        bp[i+2] = (byte)s.charAt(j+2);
        bp[i+3] = (byte)255;
      }
    }
    g.bindTexture(g.C3DL_TEXTURE_2D, idx);
    g.texImage2D(g.C3DL_TEXTURE_2D, 0,
                 g.C3DL_RGBA8, w, w, 0, g.C3DL_RGBA, g.C3DL_UNSIGNED_BYTE,
                 bp);
  }

  private void makeTexture() {
    makeImageTexture(0, Dt.et, 32);
    makeImageTexture(1, Dt.sp, 32);
    makeImageTexture(2, Dt.st, 16);
    makeImageTexture(3, Dt.sc, 4);

    for ( int i=0 ; i<5 ; i++ ) {
      makeStringTexture(4+i*2, bb[i][0], false);
      makeStringTexture(5+i*2, bb[i][1], true);
    }

    for ( int i=0 ; i<10 ; i++ ) {
      makeStringTexture(20+i, String.valueOf(i), false);
    }
    makeStringTexture(30, "score", false);
  }

  // Viewpoint data.
  private int ex, ey, ez, eax, eay, eaz;
  private int sx, sy, sz, sax, say, saz;

  // Game status (0-title  1-in game 2-crashed 3-gamvover).
  private int gs;
  private int gc;
  private int sc;

  // Meteo data.
  private final int MM = 16;
  private int[] mx = new int[MM];
  private int[] my = new int[MM];
  private int[] mz = new int[MM];
  private int[] mh = new int[MM];
  private int[] mmx = new int[MM];
  private int[] mmy = new int[MM];
  private int[] mmz = new int[MM];
  private int mi = MM;

  // Scope data.
  private int sd1, sd2;
  private int scx, scy, scz;

  // Laser data.
  private final int LM = 4;
  private int[] lx = new int[LM];
  private int[] ly = new int[LM];
  private int[] lz = new int[LM];
  private int[] lpx = new int[LM];
  private int[] lpy = new int[LM];
  private int[] lpz = new int[LM];
  private int[] lc = new int[LM];
  private int[] lmx = new int[LM];
  private int[] lmy = new int[LM];
  private int[] lmz = new int[LM];
  private int li = LM;
  private int fc, fsc = 0;

  // Game rank.
  private int rk;
  // Damage and score.
  private int dm, scr;

  // A width and the height of the earth.
  private final int EW = 640*fx1, EH = 64*fx1;
  // A width of the meteo.
  private final int MW = fx1*16;
  // An initial height of the meteo.
  private final int MH = EH*3;

  /**
   * Initialize
   */
  private void initGameState() {
    sax = 0; say = 0; saz = -EW/3;
    eax = 0; eay = 0; eaz = saz-fx1;
    for ( int i=0 ; i<MM ; i++ ) {
      mh[i] = -1;
    }
    sd1 = sd2 = 0;
    for ( int i=0 ; i<LM ; i++ ) {
      lc[i] = -1;
    }
    fc = 2;
    rk = 512;
    dm = 0;
  }

  /**
   * Start title.
   */
  private void initTitle() {
    gs = 0; gc = -1;
    initGameState();
  }

  /**
   * Start game.
   */
  private void initInGame() {
    gs = 1; gc = -1;
    scr = 0;
    initGameState();
  }

  /**
   * Start gameover.
   */
  private void initGameover() {
    gs = 3; gc = -1;
  }

  /**
   * Move an eye position and a looking position.
   */
  private void moveViewpoint() {
    ex += (eax-ex)>>2; ey += (eay-ey)>>2; ez += (eaz-ez)>>2;
    sx += (sax-sx)>>4; sy += (say-sy)>>4; sz += (saz-sz)>>4;
  }

  // Title and gameover messages.
  private final String bb[][] = {
    {"laser", "command"},
    {"push[7]", "to start"},
    {"[arrow]", "scope"},
    {"[7]", "fire"},
    {"  game", "over  "}
  };

  private int dx, dy, dd, dt;
  /**
   * Set the billboard for title and gameover.
   */
  private void moveBillboard() {
    if ( (gc%92) == 0 ) {
      int mc = (gc/92)%4;
      if ( gs == 3 ) {
        if ( gc > 0 ) return;
        mc = 4;
      }
      dt = 4+mc*2;
      dx = 0; dy = 0;

      saz = dy;
      say = eay = fx1/16;
      if ( dd == 0 ) {
        dd = 1;
        eax = sax = dx+23500;
        eaz = dy + fx1*4/3;
      } else {
        dd = 0;
        eax = sax = dx+9000;
        eaz = dy - fx1*4/3;
      }
    } else {
      if ( dd == 0 ) {
        eaz += fx1/372;
      } else {
        eaz -= fx1/372;
      }
    }
  }

  /*
   * Move meteos and lasers.
   */

  private void addMeteo() {
    int i;
    for ( i=0 ; i<MM ; i++ ) {
      mi--; if ( mi < 0 ) mi += MM;
      if ( mh[mi] < 0 ) break;
    }
    if ( i >= MM ) return;
    mx[mi] = randNS(EW/6);
    my[mi] = MH;
    mz[mi] = EW/4 + randNS(EW/10);
    int tn = 20*(10240+randNS(4800))/rk;
    mmy[mi] = -MH*2/tn;
    mmx[mi] = (randNS(EW/6)-mx[i])/tn;
    mmz[mi] = randNS(EW/8/tn);
    mh[mi] = 0;
  }

  private void addLaser() {
    int i;
    for ( i=0 ; i<LM ; i++ ) {
      li--; if ( li < 0 ) li += LM;
      if ( lc[li] < 0 ) break;
    }
    if ( i >= LM ) return;
    if ( (fsc&1) == 0 ) {
      lx[li] = ex - fx1*8;
    } else {
      lx[li] = ex + fx1*8;
    }
    fsc++;
    ly[li] = ey - fx1*4;
    lz[li] = ez;
    lmx[li] = (scx-lx[li]) / 2;
    lmy[li] = (scy-ly[li]) / 2;
    lmz[li] = (scz-lz[li]) / 2;
    lc[li] = 6;
  }

  private void moveMeteos() {
    int mny = MH*2;
    for ( int i=0 ; i<MM ; i++ ) {
      if ( mh[i] < 0 ) continue;
      mx[i] += mmx[i];
      my[i] += mmy[i];
      mz[i] += mmz[i];
      if ( mh[i] > 0 ) {
        mh[i]++;
        if ( mh[i] == 8 || mh[i] == 16 ) {
          mh[i] = -1;
          continue;
        }
      } else if ( my[i] < -(MH/5)*4 ) {
        mh[i] = 8;
        for ( int j=0 ; j<MM ; j++ ) {
          if ( mh[j] >= 0 && j != i ) {
            mh[j] = -1;
          }
        }
        rk -= (rk>>2);
        dm++;
        if ( dm >= 3 ) {
          initGameover();
        }
        continue;
      }
      if ( gs == 1 && mny > my[i] ) {
        mny = my[i];
        sax = mx[i]; say = my[i]; saz = mz[i];
      }
    }
  }

  private void moveLasers() {
    for ( int i=0 ; i<LM ; i++ ) {
      if ( lc[i] < 0 ) continue;
      lpx[i] = lx[i];
      lpy[i] = ly[i];
      lpz[i] = lz[i];
      lx[i] += lmx[i];
      ly[i] += lmy[i];
      lz[i] += lmz[i];
      lc[i]--;
      for ( int j=0 ; j<MM ; j++ ) {
        if ( mh[j] != 0 ) continue;
        if ( Math.abs(lx[i]-mx[j]) + Math.abs(ly[i]-my[j]) +
         Math.abs(lz[i]-mz[j]) < fx1*64 ) {
          mh[j] = 1;
          scr++;
        }
      }
    }
  }

  private int d = 0;

  /**
   * Move 1 frame.
   */
  private void move() {
    switch ( gs ) {
    case 0:
      moveBillboard();
      break;
    case 1:
      if ( randN(24) == 0 ) addMeteo();
      moveMeteos();
      moveLasers();
      rk++;
      break;
    case 3:
      moveMeteos();
      moveBillboard();
      if ( gc > 160 ) {
        initTitle();
        break;
      }
      break;
    }
    if ( gc >= 0 ) moveViewpoint();
    gc++;
  }

  /**
   * Adjust a scope's position.
   */
  private void checkScopePos() {
    if ( sz-ez < fx1 ) return;
    int six = sx*160/(sz-ez), siy = sy*160/(sz-ez);
    if ( sd1 < -50+six ) sd1 = -50+six;
    else if ( sd1 > 50+six ) sd1 = 50+six;
    if ( sd2 < -50+siy ) sd2 = -50+siy;
    else if ( sd2 > 50+siy ) sd2 = 50+siy;
  }

  private int kd = 4;

  private boolean pPrsd = true, sPrsd = true;

  /**
   * Check a key pad state.
   */
  private void handleKeyState() {
    int key = canvas.getKeypadState();
    if ( (key&(1<<Display.KEY_SOFT2)) != 0 ) {
      terminate();
    }
    if ( (key&(1<<Display.KEY_POUND)) == 0 ) {
      pPrsd = false;
    } else if ( !pPrsd ) {
      pPrsd = true;
      if ( backLight ) {
        backLight = false;
      } else {
        backLight = true;
      }
      setBackLight();
    }
    switch ( gs ) {
    case 0:
      if ( (key&(1<<Display.KEY_7)) == 0 ) {
        sPrsd = false;
      } else if ( !sPrsd ) {
        sPrsd = true;
        initInGame();
      }
      break;
    case 1:
      if ( (key&((1<<Display.KEY_UP) | (1<<Display.KEY_UPPER_LEFT) |
       (1<<Display.KEY_UPPER_RIGHT)) ) != 0 ) {
        sd1 += kd;
      }
      if ( (key&((1<<Display.KEY_DOWN) | (1<<Display.KEY_LOWER_LEFT) |
       (1<<Display.KEY_LOWER_RIGHT)) ) != 0 ) {
        sd1 -= kd;
      }
      if ( (key&((1<<Display.KEY_RIGHT) | (1<<Display.KEY_UPPER_RIGHT) |
       (1<<Display.KEY_LOWER_RIGHT)) ) != 0 ) {
        sd2 += kd;
      }
      if ( (key&((1<<Display.KEY_LEFT) | (1<<Display.KEY_UPPER_LEFT) |
       (1<<Display.KEY_LOWER_LEFT)) ) != 0 ) {
        sd2 -= kd;
      }
      checkScopePos();
      if ( (key&(1<<Display.KEY_7)) != 0 ) {
        if ( fc <= 0 ) {
          fc = 2;
          addLaser();
        } else {
          fc--;
        }
        kd = 2;
      } else {
        fc = 0;
        kd = 5;
      }
      break;
    case 2:
      if ( (key&(1<<Display.KEY_7)) == 0 ) {
        sPrsd = false;
      } else if ( !sPrsd ) {
        sPrsd = true;
        initGameover();
      }
      break;
    case 3:
      if ( (key&(1<<Display.KEY_7)) == 0 ) {
        sPrsd = false;
      } else if ( !sPrsd ) {
        sPrsd = true;
        initTitle();
      }
      break;
    }
  }

  /*
   * Draw a scope, meteos, lasers and the earth.
   */

  private final int SW = 2*fx1;

  private void drawScope() {
    g.bindTexture(g.C3DL_TEXTURE_2D, 3);
    g.pushMatrix();
    scx = sd1*fx1; scy = sd2*fx1; scz = ez+fx1*128;
    g.translate(scx, scy, scz);
    g.begin(g.C3DL_QUADS);
    g.texCoord(0, 0);
    g.vertex(-SW, -SW, 0);
    g.texCoord(fx1, 0);
    g.vertex( SW, -SW, 0);
    g.texCoord(fx1, fx1);
    g.vertex( SW,  SW, 0);
    g.texCoord(0, fx1);
    g.vertex(-SW,  SW, 0);
    g.end();
    g.popMatrix();
  }

  private void drawMeteos() {
    g.bindTexture(g.C3DL_TEXTURE_2D, 2);
    for ( int i=0 ; i<MM ; i++ ) {
      if ( mh[i] < 0 ) continue;
      g.pushMatrix();
      g.translate(mx[i], my[i], mz[i]);
      g.rotate(my[i]*45/MH, fx1, 0, 0);
      if ( mh[i] == 0 ) {
        g.color(fx1, fx1/3*2, fx1/3*2, fx1);
      } else {
        if ( mh[i] < 8 ) {
          int c = mh[i]*(fx1/8);
          g.color(c, c, c, c);
        } else {
          g.color(fx1, 0, 0, fx1);
        }
      }
      g.begin(g.C3DL_QUADS);
      g.texCoord(0, 0);
      g.vertex(-MW, -MW, 0);
      g.texCoord(fx1, 0);
      g.vertex( MW, -MW, 0);
      g.texCoord(fx1, fx1);
      g.vertex( MW,  MW, 0);
      g.texCoord(0, fx1);
      g.vertex(-MW,  MW, 0);
      g.end();
      int z;
      if ( mh[i] > 0 ) {
        if ( mh[i] < 8 ) {
          z = mh[i]*fx1*2;
          int c = (8-mh[i])*(fx1/8);
          g.color(c, c, c, c);
        } else {
          z = (mh[i]-8)*fx1*5;
          int c = (16-mh[i])*(fx1/8);
          g.color(fx1, c, c, fx1);
        }
      } else if ( my[i] < 0 ) {
        int r = (int)(-(long)my[i]*fx1/MH);
        z = randNS(fx1/2) + fx1/2 + r*3;
        g.color(fx1, fx1-r, fx1-r, fx1);
      } else {
        z = randNS(fx1/2) + fx1/2;
        g.color(fx1, fx1, fx1, fx1);
      }
      g.scale(z, z, z);
      g.rotate(randN(360), 0, 0, fx1);
      g.begin(g.C3DL_QUADS);
      g.texCoord(0, 0);
      g.vertex(-MW, -MW, 0);
      g.texCoord(fx1, 0);
      g.vertex( MW, -MW, 0);
      g.texCoord(fx1, fx1);
      g.vertex( MW,  MW, 0);
      g.texCoord(0, fx1);
      g.vertex(-MW,  MW, 0);
      g.end();
      g.popMatrix();
    }
  }

  private void drawLasers() {
    g.disable(g.C3DL_TEXTURE_2D);
    g.color(fx1, fx1, fx1/3*2, fx1);
    g.begin(g.C3DL_LINES);
    for ( int i=0 ; i<LM ; i++ ) {
      if ( lc[i] < 0 ) continue;
      g.vertex(lx[i], ly[i], lz[i]);
      g.vertex(lpx[i], lpy[i], lpz[i]);
    }
    g.end();
    g.enable(g.C3DL_TEXTURE_2D);
  }

  private void drawEarth() {
    g.bindTexture(g.C3DL_TEXTURE_2D, 0);
    g.color(fx1/3*2, fx1*2/(dm+3), fx1*3/(dm+3), fx1);
    g.begin(g.C3DL_QUADS);
    g.texCoord(0, 0);
    g.vertex(-EW, -EH,  EW);
    g.texCoord(fx1, 0);
    g.vertex( EW, -EH,  EW);
    g.texCoord(fx1, fx1);
    g.vertex( EW, -EH, -EW);
    g.texCoord(0, fx1);
    g.vertex(-EW, -EH, -EW);
    g.end();
    g.bindTexture(g.C3DL_TEXTURE_2D, 1);
    g.color(fx1, fx1, fx1, fx1);
    g.begin(g.C3DL_QUADS);
    g.texCoord(0, 0);
    g.vertex(-EW,  EH,  EW);
    g.texCoord(fx1, 0);
    g.vertex( EW,  EH,  EW);
    g.texCoord(fx1, fx1);
    g.vertex( EW,  EH, -EW);
    g.texCoord(0, fx1);
    g.vertex(-EW,  EH, -EW);
    g.end();
  }

  /**
   * Draw a billboard for a title/gameover.
   */
  private void drawBillboard() {
    g.color(fx1, fx1, fx1, fx1);

    switch ( dd ) {
    case 0:
      g.bindTexture(g.C3DL_TEXTURE_2D, dt);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(fx1-2560, 0);
      g.vertex(dx, fx1/16*2, dy);
      g.texCoord(0, 0);
      g.vertex(dx+fx1/2, fx1/16*2, dy);
      g.texCoord(0, fx1-2560);
      g.vertex(dx+fx1/2, fx1/16, dy);
      g.texCoord(fx1-2560, fx1-2560);
      g.vertex(dx, fx1/16, dy);
      g.end();
      g.bindTexture(g.C3DL_TEXTURE_2D, dt+1);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(fx1-2560, 0);
      g.vertex(dx-fx1/4, fx1/16, dy);
      g.texCoord(0, 0);
      g.vertex(dx-fx1/4+fx1/2, fx1/16, dy);
      g.texCoord(0, fx1-2560);
      g.vertex(dx-fx1/4+fx1/2, 0, dy);
      g.texCoord(fx1-2560, fx1-2560);
      g.vertex(dx-fx1/4, 0, dy);
      g.end();
      break;
    case 1:
      g.bindTexture(g.C3DL_TEXTURE_2D, dt);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(0, 0);
      g.vertex(dx, fx1/16*2, dy);
      g.texCoord(fx1-2560, 0);
      g.vertex(dx+fx1/2, fx1/16*2, dy);
      g.texCoord(fx1-2560, fx1-2560);
      g.vertex(dx+fx1/2, fx1/16, dy);
      g.texCoord(0, fx1-2560);
      g.vertex(dx, fx1/16, dy);
      g.end();
      g.bindTexture(g.C3DL_TEXTURE_2D, dt+1);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(0, 0);
      g.vertex(dx+fx1/4, fx1/16, dy);
      g.texCoord(fx1-2560, 0);
      g.vertex(dx+fx1/4+fx1/2, fx1/16, dy);
      g.texCoord(fx1-2560, fx1-2560);
      g.vertex(dx+fx1/4+fx1/2, 0, dy);
      g.texCoord(0, fx1-2560);
      g.vertex(dx+fx1/4, 0, dy);
      g.end();
      break;
    }
  }

  /**
   * Draw a score.
   */
  private void drawScore() {
    g.color(fx1, fx1, fx1, fx1);

    int x = fx1/18*6, s = scr;

    do {
      g.bindTexture(g.C3DL_TEXTURE_2D, s%10+20);
      g.begin(g.C3DL_TRIANGLE_FAN);
      g.texCoord(0, 0);
      g.vertex(x, fx1/16*6, -fx1);
      g.texCoord(fx1-2560, 0);
      g.vertex(x+fx1/16*6/8, fx1/16*6, -fx1);
      g.texCoord(fx1-2560, fx1-2560);
      g.vertex(x+fx1/16*6/8, fx1/16*4, -fx1);
      g.texCoord(0, fx1-2560);
      g.vertex(x, fx1/16*4, -fx1);
      g.end();
      s /= 10; x -= fx1/16*6/8;
    } while ( s > 0 );

    g.bindTexture(g.C3DL_TEXTURE_2D, 30);
    g.begin(g.C3DL_TRIANGLE_FAN);
    g.texCoord(0, 0);
    g.vertex(-fx1/16*6, fx1/16*6, -fx1);
    g.texCoord(fx1-2560, 0);
    g.vertex(0, fx1/16*6, -fx1);
    g.texCoord(fx1-2560, fx1-2560);
    g.vertex(0, fx1/16*4, -fx1);
    g.texCoord(0, fx1-2560);
    g.vertex(-fx1/16*6, fx1/16*4, -fx1);
    g.end();
  }

  /**
   * Draw to the canvas.
   */
  private void draw() {
    g.set3DMode(true);

    g.clear(g.C3DL_COLOR_BUFFER);

    g.pushMatrix();
    g.rotate(-90*fx1, 0, 0, fx1);

    // Draw front billboards.
    if ( gs != 0 ) {
      drawScore();
    }

    g.lookAt(ex, ey, ez, sx, sy, sz,
             -sin(0<<3), cos(0<<3), 0);

    switch (gs) {
    case 0:
      drawEarth();
      drawBillboard();
      break;
    case 1:
      drawEarth();
      drawMeteos();
      drawLasers();
      drawScope();
      break;
    case 3:
      drawEarth();
      drawMeteos();
      drawBillboard();
      break;
    }

    g.popMatrix();
    g.flush();
  }

  /**
   * Main loop. (20 frames/sec)
   */
  public void run() {
    long prvTicks, nowTicks;
    int frame;
    prvTicks = System.currentTimeMillis();
    for ( ; ; ) {
      nowTicks = System.currentTimeMillis();
      frame = (int)(nowTicks-prvTicks)/50;
      if ( frame <= 0 ) {
        frame = 1;
        prvTicks += 50;
        while ( System.currentTimeMillis() < prvTicks ) {}
      } else {
        prvTicks = nowTicks;
        if ( frame > 4 ) {
          frame = 4;
        }
      }
      if ( g == null ) {
        g = canvas.getC3DGraphics();
        if ( g != null ) {
          initCanvas();
        } else {
          continue;
        }
      }
      for ( ; frame>0 ; frame-- ) {
        handleKeyState();
        move();
      }
      g.lock();
      draw();
      g.unlock(true);
    }
  }

  /*
   * Random util methods.
   */
  private static Random rnd = new Random();

  private int randN(int n) {
    return Math.abs(rnd.nextInt()%n);
  }

  private int randNS(int n) {
    return rnd.nextInt()%n;
  }

  /*
   * SinCos util methods.
   */
  private int[] sctbl = new int[1024+256];

  private void initSinCos() {
    for ( int i=0 ; i<1024+256 ; i++ ) {
      sctbl[i] = C3DMath.sin(i*(360*fx1/1024));
    }
  }

  private int sin(int d) {
    return sctbl[((d>>6)/360)&1023];
  }

  private int cos(int d) {
    return sctbl[(((d>>6)/360)&1023) + 256];
  }

  /**
   * Bootstrap.
   */
  public void start() {
    initSinCos();
    init();
    run();
  }
}
